﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace TestThreadPool
{
    public partial class Form1 : Form
    {
        // Maximalzahl an zu startenden Threads (System-limitiert)
        private const int MaxNbThreads = 64;

        // Dummy-Variable, die der Compiler nicht wegoptimiert
        public double dummy;

        // zur internen Thread-synchronisation
        private ManualResetEvent[] doneEvents = new ManualResetEvent[MaxNbThreads];

        // auf Start oder Beendigung aller Threads triggern
        private bool SendStartEvent;

        // Bit-Flag für ProcessorAffinity
        private int FlagUsedCores;

        private Stopwatch duration;

        public Form1()
        {
            this.InitializeComponent();

            this.numericUpDown1.Minimum = 1;
            this.numericUpDown1.Maximum = Environment.ProcessorCount;
            this.numericUpDown1.Value = Environment.ProcessorCount - 1;

            this.progressBar1.Maximum = MaxNbThreads;

            this.duration = new Stopwatch();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            this.textBox1.Clear();
            this.progressBar1.Value = 0;
            this.SendStartEvent = false;

            // für jeden zu benutzenden Kern ein Bit setzen
            this.FlagUsedCores = (int)Math.Pow(2, (double)this.numericUpDown1.Value) - 1;

            this.duration.Restart();
            this.backgroundWorker1.RunWorkerAsync();
        }

        private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
        {
            // Anzahl der zu nutzenden Prozessor-Kerne vorgeben
            Process.GetCurrentProcess().ProcessorAffinity = new IntPtr(this.FlagUsedCores);

            // praktisch kein Einfluss
            //ThreadPool.SetMaxThreads(2, 2);

            // alle Threads starten
            for (int i = 0; i < MaxNbThreads; i++)
            {
                // Threads dem ThreadPool hinzufügen, initial keine Meldung
                ThreadPool.QueueUserWorkItem(new WaitCallback(DoWork), i);
                this.doneEvents[i] = new ManualResetEvent(false);
                this.backgroundWorker1.ReportProgress(i);
            }

            // Warten auf den Start oder Beenden aller Threads
            WaitHandle.WaitAll(this.doneEvents);

            this.backgroundWorker1.ReportProgress(MaxNbThreads);
            // Testausgabe Feddich
            this.LogText("@@@@@@@@@@@@@@@@@@@@@@@@", true);
        }

        public void DoWork(object o)
        {
            // Laufvariable holen
            int i = Convert.ToInt32(o);

            if (this.SendStartEvent)
            {
                // Hier wird das Starten der Threads gemeldet
                this.doneEvents[i].Set();
            }

            // Testausgabe Start Thread
            this.LogText(string.Format("start {0} - ", i), false);

            // CPU-Auslastung in die Höhe treiben
            for (int j = 0; j < 10000000; j++)
            {
                this.dummy = Math.Atan2(i, i + 100);
            }

            // Testausgabe End Thread
            this.LogText(string.Format("end {0}", i), true);

            if (!this.SendStartEvent)
            {
                // Hier wird das Beenden der Threads gemeldet
                this.doneEvents[i].Set();
            }
        }

        private void LogText(string txt, bool newline)
        {
            //Console.Write(txt);
            //if (newline)
            //{
            //    Console.WriteLine("");
            //}
            //return;

            if (this.textBox1.InvokeRequired)
            {
                this.Invoke(new Action<string, bool>(this.LogText), txt, newline);
                return;
            }

            this.textBox1.AppendText(txt);
            if (newline)
            {
                this.textBox1.AppendText(Environment.NewLine);
            }

            this.textBox1.Refresh();
        }

        private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            this.progressBar1.Value = e.ProgressPercentage;
        }

        private void backgroundWorker1_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            this.duration.Stop();
            this.LogText("", true); // Leerzeile
            this.LogText(string.Format("Duration: {0:0.000} s", this.duration.ElapsedMilliseconds / 1000.0), true);
        }
    }
}
